import os, sys
import unittest
import subprocess
import shutil
import logging
import io
import stat
from pathlib import Path as _Path
from multiprocessing import freeze_support

from clang_build import clang_build
from clang_build.errors import CompileError
from clang_build.errors import LinkError

def on_rm_error( func, path, exc_info):
    # path contains the path of the file that couldn't be removed
    # let's just assume that it's read-only and unlink it.
    os.chmod( path, stat.S_IWRITE )
    os.unlink( path )

def clang_build_try_except( args ):
    try:
        clang_build.build(clang_build.parse_args(args))
    except CompileError as compile_error:
        logger = logging.getLogger('clang_build')
        logger.error('Compilation was unsuccessful:')
        for target, errors in compile_error.error_dict.items():
            printout = f'Target {target} did not compile. Errors:'
            for file, output in errors:
                for out in output:
                    row = out['row']
                    column = out['column']
                    messagetype = out['type']
                    message = out['message']
                    printout += f'\n{file}:{row}:{column}: {messagetype}: {message}'
            logger.error(printout)
    except LinkError as link_error:
        logger = logging.getLogger('clang_build')
        logger.error('Linking was unsuccessful:')
        for target, errors in link_error.error_dict.items():
            printout = f'Target {target} did not link. Errors:\n{errors}'
            logger.error(printout)

class TestClangBuild(unittest.TestCase):
    def test_hello_world_mwe(self):
        clang_build_try_except(['-d', 'test/mwe', '-p'])

        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'Hello!')

    def test_compile_error(self):
        with self.assertRaises(CompileError):
            clang_build.build(clang_build.parse_args(['-d', 'test/mwe_build_error', '-V', '-p']))

    def test_script_call(self):
        try:
            subprocess.check_output(['clang-build', '-d', 'test/mwe', '-V', '-p'], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            self.fail('Compilation failed')
        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'Hello!')

    def test_hello_world_rebuild(self):
        clang_build_try_except(['-d', 'test/mwe', '-p'])
        logger = logging.getLogger('clang_build')
        logger.setLevel(logging.DEBUG)
        stream_capture = io.StringIO()
        ch = logging.StreamHandler(stream_capture)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)
        clang_build_try_except(['-d', 'test/mwe', '-V', '-p'])
        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        logger.removeHandler(ch)

        self.assertRegex(stream_capture.getvalue(), r'.*Target \[main\] is already compiled.*')
        self.assertEqual(output, 'Hello!')

    def test_automatic_include_folders(self):
        clang_build_try_except(['-d', 'test/mwe_with_default_folders', '-V', '-p'])

        try:
            output = subprocess.check_output(['./build/default/bin/main'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'Calculated Magic: 30')

    def test_toml_mwe(self):
        clang_build_try_except(['-d', 'test/toml_mwe', '-p'])

        try:
            output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'Hello!')

    def test_toml_custom_folder(self):
        clang_build_try_except(['-d', 'test/toml_with_custom_folder', '-p'])

        try:
            output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'Hello!')

    def test_external_scripts(self):
        clang_build_try_except(['-d', 'test/external_scripts', '-V', '-p'])

        try:
            output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'the version is 1.2.0')

    def test_subproject(self):
        clang_build_try_except(['-d', 'test/subproject', '-V', '-p'])

        try:
            output = subprocess.check_output(['./build/mainproject/default/bin/runLib'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, 'Hello! mylib::triple(3) returned 9')

    def test_boost_filesystem(self):
        clang_build_try_except(['-d', 'test/boost-filesystem', '-V', '-p'])

        try:
            output = subprocess.check_output(['./build/myexe/default/bin/myexe', 'build'], stderr=subprocess.STDOUT).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            self.fail('Could not run compiled program')

        self.assertEqual(output, '"build" is a directory')

    # def test_openmp(self):
    #     clang_build_try_except(['-d', 'test/openmp', '-V', '-p'])

    #     try:
    #         output = subprocess.check_output(['./build/default/bin/runHello'], stderr=subprocess.STDOUT).decode('utf-8').strip()
    #     except subprocess.CalledProcessError:
    #         self.fail('Could not run compiled program')

        # self.assertEqual(output, 'Hello from thread 1, nthreads 8')

    # def test_mwe_two_targets(self):
    #     clang_build_try_except(['-d', 'test/multi_target_external', '-V', '-p'])

    #     try:
    #         output = subprocess.check_output(['./build/myexe/default/bin/runLib'], stderr=subprocess.STDOUT).decode('utf-8').strip()
    #     except subprocess.CalledProcessError:
    #         self.fail('Could not run compiled program')

    #     self.assertEqual(output, 'Hello!')

    def tearDown(self):
        if _Path('build').exists():
            shutil.rmtree('build', onerror = on_rm_error)


if __name__ == '__main__':
    freeze_support()
    unittest.main()


import networkx as _nx

def find_non_existent_dependencies(project):
    illegal_dependencies = []
    for nodename, node in project.items():
        for dependency in node.get('dependencies', []):
            # Split string at dots
            subnames = str(dependency).split(".")
            if subnames[0] in project.keys():
                subproject = project
                for i in range(1, len(subnames)):
                    subproject = subproject[subnames[i-1]]
                    if subnames[i] not in subproject.keys():
                        illegal_dependencies.append((nodename, '.'.join(subnames[:i])))
                        i = len(subnames)
            else:
                illegal_dependencies.append((nodename, dependency))

    return illegal_dependencies

def find_circular_dependencies(project):
    graph = _nx.DiGraph()
    for nodename, node in project.items():
        for dependency in node.get('dependencies', []):
            graph.add_edge(str(nodename), str(dependency))

    return list(_nx.simple_cycles(graph))

def get_dependency_walk(project):
    graph = _nx.DiGraph()
    for nodename, node in project.items():
        dependencies = node.get('dependencies', [])
        if not dependencies:
            graph.add_node(str(nodename))
            continue

        for dependency in dependencies:
            # Split string at dots
            subnames = str(dependency).split(".")
            graph.add_edge(str(nodename), str(subnames[-1]))

    return list(reversed(list(_nx.topological_sort(graph))))


'''
Target describes a single build or dependency target with all needed paths and
a list of buildables that comprise it's compile and link steps.
'''

import os as _os
import textwrap as _textwrap
import sys
from pathlib import Path as _Path
import subprocess as _subprocess
from multiprocessing import freeze_support as _freeze_support
import logging as _logging

import toml
from .dialect_check import get_max_supported_compiler_dialect as _get_max_supported_compiler_dialect
from .build_type import BuildType as _BuildType
from .target import Executable as _Executable,\
                    SharedLibrary as _SharedLibrary,\
                    StaticLibrary as _StaticLibrary,\
                    HeaderOnly as _HeaderOnly
from .dependency_tools import find_circular_dependencies as _find_circular_dependencies,\
                              find_non_existent_dependencies as _find_non_existent_dependencies,\
                              get_dependency_walk as _get_dependency_walk
from .io_tools import get_sources_and_headers as _get_sources_and_headers
from .progress_bar import CategoryProgress as _CategoryProgress,\
                          IteratorProgress as _IteratorProgress
from .logging_stream_handler import TqdmHandler as _TqdmHandler

_LOGGER = _logging.getLogger('clang_build.clang_build')



class Project:
    def __init__(self, config, environment, multiple_projects):

        self.name = config.get("name", "")

        self.working_directory = environment.working_directory

        # Project build directory
        self.build_directory = environment.build_directory
        if multiple_projects:
            self.build_directory = self.build_directory.joinpath(self.name)

        self.external = "url" in config
        if self.external:
            download_directory = self.build_directory.joinpath('external_sources')
            # Check if directory is already present and non-empty
            if download_directory.exists() and _os.listdir(str(download_directory)):
                _LOGGER.info(f'External project [[{self.name}]]: sources found in {str(download_directory)}')
            # Otherwise we download the sources
            else:
                _LOGGER.info(f'External project [[{self.name}]]: downloading to {str(download_directory)}')
                download_directory.mkdir(parents=True, exist_ok=True)
                try:
                    _subprocess.run(["git", "clone", config["url"], str(download_directory)], stdout=_subprocess.PIPE, stderr=_subprocess.PIPE, encoding='utf-8')
                except _subprocess.CalledProcessError as e:
                    error_message = f"Error trying to download external project [[{self.name}]]. Message " + e.output
                    _LOGGER.exception(error_message)
                    raise RuntimeError(error_message)
                _LOGGER.info(f'External project [[{self.name}]]: downloaded')
            self.working_directory = download_directory

        if "directory" in config:
            self.working_directory = environment.working_directory.joinpath(config["directory"])
            toml_file = _Path(self.working_directory, 'clang-build.toml')
            if toml_file.exists():
                environment.logger.info(f'Found config file {toml_file}')
                config = toml.load(str(toml_file))
            else:
                error_message = f"Project {self.name}: could not find project file in directory {self.working_directory}"
                _LOGGER.exception(error_message)
                raise RuntimeError(error_message)

        # Get subset of config which contains targets not associated to any project name
        self.targets_config = {key: val for key, val in config.items() if not key == "subproject" and not key == "name"}

        # Get subsets of config which define projects
        self.subprojects_config = {key: val for key, val in config.items() if key == "subproject"}

        # An "anonymous" project, i.e. project-less targets, is not allowed together with subprojects
        if self.targets_config and self.subprojects_config:
            if not self.name:
                error_message = f"Project {self.name}: Your config file specified one or more projects. In this case you are not allowed to specify targets which do not belong to a project."
                _LOGGER.exception(error_message)
                raise RuntimeError(error_message)

        # Generate Projects
        subprojects = []
        # if targets_config:
        #     subprojects += [Project(targets_config, environment, multiple_projects)]

        if self.subprojects_config:
            subprojects += [Project(config, environment, multiple_projects) for config in self.subprojects_config["subproject"]]

        self.subprojects = subprojects
        self.subproject_names = [project.name if project.name else "anonymous" for project in subprojects]

        # Use sub-build directories if the project contains multiple targets
        multiple_targets = False
        if len(self.targets_config.items()) > 1:
            multiple_targets = True

        if not self.targets_config:
            return

        targets_and_subprojects = self.targets_config.copy()
        for project in subprojects:
            targets_and_subprojects[project.name] = project.targets_config

        targets_and_subproject_targets = self.targets_config.copy()
        for project in subprojects:
            targets_and_subproject_targets.update(project.targets_config)

        # Parse targets from toml file
        non_existent_dependencies = _find_non_existent_dependencies(targets_and_subprojects)
        if non_existent_dependencies:
            error_messages = [f'In {target}: the dependency {dependency} does not point to a valid target' for\
                            target, dependency in non_existent_dependencies]

            error_message = _textwrap.indent('\n'.join(error_messages), prefix=' '*3)
            _LOGGER.exception(error_message)
            raise RuntimeError(error_message)

        circular_dependencies = _find_circular_dependencies(self.targets_config)
        if circular_dependencies:
            error_messages = [f'In {target}: circular dependency -> {dependency}' for\
                            target, dependency in circular_dependencies]

            error_message = _textwrap.indent('\n'.join(error_messages), prefix=' '*3)
            _LOGGER.exception(error_message)
            raise RuntimeError(error_message)


        target_names_total = _get_dependency_walk(targets_and_subproject_targets)
        target_names_project = []
        for name in target_names_total:
            if name in self.targets_config:
                target_names_project.append(name)

        self.target_list = []


        for target_name in _IteratorProgress(target_names_project, environment.progress_disabled, len(target_names_project)):
            target_node = targets_and_subproject_targets[target_name]
            # Directories
            target_build_directory = self.build_directory if not multiple_targets else self.build_directory.joinpath(target_name)
            target_root_directory  = self.working_directory

            # If target is marked as external, try to fetch the sources
            ### TODO: external sources should be fetched before any sources are read in, i.e. even before the first target is created
            external = "url" in target_node
            if external:
                download_directory = target_build_directory.joinpath('external_sources')
                # Check if directory is already present and non-empty
                if download_directory.exists() and _os.listdir(str(download_directory)):
                    _LOGGER.info(f'External target [{target_name}]: sources found in {str(download_directory)}')
                # Otherwise we download the sources
                else:
                    _LOGGER.info(f'External target [{target_name}]: downloading to {str(download_directory)}')
                    download_directory.mkdir(parents=True, exist_ok=True)
                    try:
                        _subprocess.run(["git", "clone", target_node["url"], str(download_directory)], stdout=_subprocess.PIPE, stderr=_subprocess.PIPE, encoding='utf-8')
                    except _subprocess.CalledProcessError as e:
                        error_message = f"Error trying to download external target [{target_name}]. Message " + e.output
                        _LOGGER.exception(error_message)
                        raise RuntimeError(error_message)
                    _LOGGER.info(f'External target [{target_name}]: downloaded')
                # self.includeDirectories.append(download_directory)
                target_root_directory = download_directory

                if "version" in target_node:
                    version = target_node["version"]
                    try:
                        _subprocess.run(["git", "checkout", version], cwd=target_root_directory, stdout=_subprocess.PIPE, stderr=_subprocess.PIPE, encoding='utf-8')
                    except _subprocess.CalledProcessError as e:
                        error_message = f"Error trying to checkout target [{target_name}] version \'{version}\'. Message " + e.output
                        _LOGGER.exception(error_message)
                        raise RuntimeError(error_message)

            # Build directory for obj, bin etc. should be under build type folder, e.g. default
            target_build_directory = target_build_directory.joinpath(environment.buildType.name.lower())

            # Sub-directory, if specified
            if 'directory' in target_node:
                target_root_directory = target_root_directory.joinpath(target_node['directory'])

            # Sources
            files = _get_sources_and_headers(target_node, target_root_directory, target_build_directory)
            # Dependencies
            dependencies = []
            for name in target_node.get('dependencies', []):
                subnames = name.split(".")
                if subnames[0] in self.subproject_names:
                    idx = self.subproject_names.index(subnames[0])
                    # TODO: so far, we are only going one layer deep... this is not enough
                    subproject = self.subprojects[idx]
                    for i in range(1, len(subnames)):
                        subproject = self.subprojects[i-1]
                        for target in subproject.target_list:
                            if subnames[-1] == target.name:
                                dependencies.append(target)
                                i = len(subnames)
                if name in target_names_project:
                    dependencies.append(self.target_list[target_names_project.index(name)])

            executable_dependencies = [target for target in dependencies if target.__class__ is _Executable]

            if executable_dependencies:
                exelist = ', '.join([f'[{dep.name}]' for dep in executable_dependencies])
                environment.logger.error(f'Error: The following targets are linking dependencies but were identified as executables:\n    {exelist}')

            if 'target_type' in target_node:
                #
                # Add an executable
                #
                if target_node['target_type'].lower() == 'executable':
                    self.target_list.append(
                        _Executable(
                            target_name,
                            target_root_directory,
                            target_build_directory,
                            files['headers'],
                            files['include_directories'],
                            files['sourcefiles'],
                            environment.buildType,
                            environment.clangpp,
                            target_node,
                            dependencies))

                #
                # Add a shared library
                #
                if target_node['target_type'].lower() == 'shared library':
                    self.target_list.append(
                        _SharedLibrary(
                            target_name,
                            target_root_directory,
                            target_build_directory,
                            files['headers'],
                            files['include_directories'],
                            files['sourcefiles'],
                            environment.buildType,
                            environment.clangpp,
                            target_node,
                            dependencies))

                #
                # Add a static library
                #
                elif target_node['target_type'].lower() == 'static library':
                    self.target_list.append(
                        _StaticLibrary(
                            target_name,
                            target_root_directory,
                            target_build_directory,
                            files['headers'],
                            files['include_directories'],
                            files['sourcefiles'],
                            environment.buildType,
                            environment.clangpp,
                            environment.clang_ar,
                            target_node,
                            dependencies))

                #
                # Add a header-only
                #
                elif target_node['target_type'].lower() == 'header only':
                    if files['sourcefiles']:
                        environment.logger.info(f'Source files found for header-only target {target_name}. You may want to check your build configuration.')
                    self.target_list.append(
                        _HeaderOnly(
                            target_name,
                            target_root_directory,
                            target_build_directory,
                            files['headers'],
                            files['include_directories'],
                            environment.buildType,
                            environment.clangpp,
                            target_node,
                            dependencies))

                else:
                    environment.logger.error(f'ERROR: Unsupported target type: {target_node["target_type"]}')

            # No target specified so must be executable or header only
            else:
                if not files['sourcefiles']:
                    environment.logger.info(f'No source files found for target {target_name}. Creating header-only target.')
                    self.target_list.append(
                        _HeaderOnly(
                            target_name,
                            target_root_directory,
                            target_build_directory,
                            files['headers'],
                            files['include_directories'],
                            environment.buildType,
                            environment.clangpp,
                            target_node,
                            dependencies))
                else:
                    environment.logger.info(f'{len(files["sourcefiles"])} source files found for target {target_name}. Creating executable target.')
                    self.target_list.append(
                        _Executable(
                            target_name,
                            target_root_directory,
                            target_build_directory,
                            files['headers'],
                            files['include_directories'],
                            files['sourcefiles'],
                            environment.buildType,
                            environment.clangpp,
                            target_node,
                            dependencies))

    def get_targets(self):
        targetlist = []
        for subproject in self.subprojects:
            targetlist += subproject.get_targets()
        targetlist += self.target_list
        return targetlist


from enum import Enum as _Enum

class BuildType(_Enum):
    Default        = 'default'
    Release        = 'release'
    #MinSizeRel     = 2
    #RelWithDebInfo = 3
    Debug          = 'debug'
    #Coverage       = 5

    def __str__(self):
        return self.value

    # Let's be case insensitive
    @classmethod
    def _missing_(cls, value):
        for item in cls:
            if item.value.lower() == value.lower():
                return item
        return super()._missing_(value)



import logging as _logging
import tqdm as _tqdm

class TqdmHandler(_logging.StreamHandler):
    def __init__(self):
        _logging.StreamHandler.__init__(self)

    def emit(self, record):
        msg = self.format(record)
        _tqdm.tqdm.write(msg)


'''
clang-build:
  TODO: module docstring...
'''

import logging as _logging
from pathlib import Path as _Path
import textwrap as _textwrap
import sys
from multiprocessing import Pool as _Pool
from multiprocessing import freeze_support as _freeze_support
import argparse
from shutil import which as _which
import toml
from pbr.version import VersionInfo as _VersionInfo

from .dialect_check import get_max_supported_compiler_dialect as _get_max_supported_compiler_dialect
from .build_type import BuildType as _BuildType
from .project import Project as _Project
from .target import Executable as _Executable,\
                    SharedLibrary as _SharedLibrary,\
                    StaticLibrary as _StaticLibrary,\
                    HeaderOnly as _HeaderOnly
from .dependency_tools import find_circular_dependencies as _find_circular_dependencies,\
                              find_non_existent_dependencies as _find_non_existent_dependencies,\
                              get_dependency_walk as _get_dependency_walk
from .io_tools import get_sources_and_headers as _get_sources_and_headers
from .progress_bar import CategoryProgress as _CategoryProgress,\
                          IteratorProgress as _IteratorProgress
from .logging_stream_handler import TqdmHandler as _TqdmHandler
from .errors import CompileError as _CompileError
from .errors import LinkError as _LinkError

_v = _VersionInfo('clang-build').semantic_version()
__version__ = _v.release_string()
version_info = _v.version_tuple



def _setup_logger(log_level=None):
    logger = _logging.getLogger(__name__)
    logger.setLevel(_logging.DEBUG)

    # create formatter _and add it to the handlers
    formatter = _logging.Formatter('%(message)s')

    # add log file handler
    fh = _logging.FileHandler('clang-build.log', mode='w')
    fh.setLevel(_logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    if log_level is not None:
        ch = _TqdmHandler()
        ch.setLevel(log_level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)




_command_line_description = (
'`clang-build` is a build system to build your C++ projects. It uses the clang '
'compiler/toolchain in the background and python as the build-system\'s scripting '
'language.\n'
'For more information please visit: https://github.com/trick-17/clang-build')

def parse_args(args):
    parser = argparse.ArgumentParser(description=_command_line_description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-V', '--verbose',
                        help='activate more detailed output',
                        action='store_true')
    parser.add_argument('-p', '--progress', help='activates a progress bar output. is overruled by -V and --debug', action='store_true')
    parser.add_argument('-d', '--directory', type=_Path,
                        help='set the root source directory')
    parser.add_argument('-b', '--build-type', choices=list(_BuildType), type=_BuildType, default=_BuildType.Default,
                        help='set the build type for this project')
    parser.add_argument('-j', '--jobs', type=int, default=1,
                        help='set the number of concurrent build jobs')
    parser.add_argument('--debug', help='activates additional debug output, overrides verbosity option.', action='store_true')
    return parser.parse_args(args=args)


def _find_clang(logger):
    clangpp = _which('clang++')
    clang_ar = _which('llvm-ar')
    if clangpp:
        llvm_root = _Path(clangpp).parents[0]
    else:
        error_message = 'Couldn\'t find clang++ executable'
        logger.error(error_message)
        raise RuntimeError(error_message)
    if not clang_ar:
        error_message = 'Couldn\'t find llvm-ar executable'
        logger.error(error_message)
        raise RuntimeError(error_message)

    logger.info(f'llvm root directory: {llvm_root}')
    logger.info(f'clang++ executable: {clangpp}')
    logger.info(f'llvm-ar executable: {clang_ar}')
    logger.info(f'Newest supported C++ dialect: {_get_max_supported_compiler_dialect(clangpp)}')

    return clangpp, clang_ar



class _Environment:
    def __init__(self, args):
        # Some defaults
        self.logger = None
        self.progress_disabled = True
        self.buildType  = None
        self.clangpp  = "clang++"
        self.clang_ar = "llvm-ar"
        # Directory this was called from
        self.calling_directory = _Path().resolve()
        # Working directory is where the project root should be - this is searched for 'clang-build.toml'
        self.working_directory = self.calling_directory

        # Verbosity
        if not args.debug:
            if args.verbose:
                _setup_logger(_logging.INFO)
            else:
                # Only file log
                _setup_logger(None)
        else:
            _setup_logger(_logging.DEBUG)

        # Progress bar
        if args.progress:
            self.progress_disabled = False

        self.logger = _logging.getLogger(__name__)
        self.logger.info(f'clang-build {__version__}')

        # Check for clang++ executable
        self.clangpp, self.clang_ar = _find_clang(self.logger)

        # Working directory
        if args.directory:
            self.working_directory = args.directory.resolve()

        if not self.working_directory.exists():
            error_message = f'ERROR: specified non-existent directory [{self.working_directory}]'
            self.logger.error(error_message)
            raise RuntimeError(error_message)

        self.logger.info(f'Working directory: {self.working_directory}')

        # Build type (Default, Release, Debug)
        self.buildType = args.build_type
        self.logger.info(f'Build type: {self.buildType.name}')

        # Multiprocessing pool
        self.processpool = _Pool(processes = args.jobs)
        self.logger.info(f'Running up to {args.jobs} concurrent build jobs')

        # Build directory
        self.build_directory = _Path('build')



def build(args):
    # Create container of environment variables
    environment = _Environment(args)

    with _CategoryProgress(['Configure', 'Compile', 'Link'], environment.progress_disabled) as progress_bar:
        target_list = []
        logger = environment.logger
        processpool = environment.processpool

        # Check for build configuration toml file
        toml_file = _Path(environment.working_directory, 'clang-build.toml')
        if toml_file.exists():
            logger.info('Found config file')

            # Parse config file
            config = toml.load(str(toml_file))

            # Determine if there are multiple projects
            targets_config = {key: val for key, val in config.items() if not key == "subproject" and not key == "name"}
            subprojects_config = {key: val for key, val in config.items() if key == "subproject"}
            multiple_projects = False
            if subprojects_config:
                if targets_config or (len(subprojects_config["subproject"]) > 1):
                    multiple_projects = True

            # Create root project
            project = _Project(config, environment, multiple_projects)

            # Get list of all targets
            target_list += project.get_targets()

            # # Generate list of all targets
            # for project in working_projects:
            #     target_list.append(project.get_targets())

        # Otherwise we try to build it as a simple hello world or mwe project
        else:
            files = _get_sources_and_headers({}, environment.working_directory, environment.build_directory)

            if not files['sourcefiles']:
                error_message = f'Error, no sources and no [clang-build.toml] found in folder: {environment.working_directory}'
                logger.error(error_message)
                raise RuntimeError(error_message)
            # Create target
            target_list.append(
                _Executable(
                    'main',
                    environment.working_directory,
                    environment.build_directory.joinpath(environment.buildType.name.lower()),
                    files['headers'],
                    files['include_directories'],
                    files['sourcefiles'],
                    environment.buildType,
                    environment.clangpp))

        # Build the targets
        progress_bar.update()

        logger.info('Compile')

        for target in _IteratorProgress(target_list, environment.progress_disabled, len(target_list), lambda x: x.name):
            target.compile(environment.processpool, environment.progress_disabled)

        # No parallel linking atm, could be added via
        # https://stackoverflow.com/a/5288547/2305545
        #

        processpool.close()
        processpool.join()

        # Check for compile errors
        errors = {}
        for target in target_list:
            if target.__class__ is not _HeaderOnly:
                if target.unsuccessful_builds:
                    errors[target.name] = [source.compile_report for source in target.unsuccessful_builds]
        if errors:
            raise _CompileError('Compilation was unsuccessful', errors)

        # Link
        progress_bar.update()
        logger.info('Link')
        for target in target_list:
            target.link()

        # Check for link errors
        errors = {}
        for target in target_list:
            if target.__class__ is not _HeaderOnly:
                if target.unsuccessful_link:
                    errors[target.name] = target.link_report
        if errors:
            raise _LinkError('Linking was unsuccessful', errors)

        progress_bar.update()
        logger.info('clang-build finished.')



def main():
    # Build
    try:
        build(parse_args(sys.argv[1:]))
    except _CompileError as compile_error:
        logger = _logging.getLogger(__name__)
        logger.error('Compilation was unsuccessful:')
        for target, errors in compile_error.error_dict.items():
            printout = f'Target [{target}] did not compile. Errors:\n'
            printout += ' '.join(errors)
            logger.error(printout)
    except _LinkError as link_error:
        logger = _logging.getLogger(__name__)
        logger.error('Linking was unsuccessful:')
        for target, errors in link_error.error_dict.items():
            printout = f'Target [{target}] did not link. Errors:\n{errors}'
            logger.error(printout)



if __name__ == '__main__':
    _freeze_support()
    main()



import os
print("before compile here ", os.getcwd())
with open("version.h", 'w') as f:
    f.write(
        "#define VERSION_MAJOR 1\n"
        "#define VERSION_MINOR 2\n"
        "#define VERSION_PATCH 0")


import subprocess as _subprocess
from functools import lru_cache as _lru_cache

_UNSUPPORTED_DIALECT_MESSAGE = "error: invalid value 'c++{0:02d}'"

def get_dialect_string(year):
    return '-std=c++{:02d}'.format(year)


def _check_dialect(year, clangpp):
    std_opt = get_dialect_string(year)
    try:
        _subprocess.run([clangpp, std_opt, '-x', 'c++', '-E', '-'],\
           check=True,input=b'',stdout=_subprocess.PIPE,stderr=_subprocess.PIPE)
    except _subprocess.CalledProcessError as e:
        error_message = _UNSUPPORTED_DIALECT_MESSAGE.format(year)
        strerr = e.stderr.decode('utf8','strict')
        if error_message in strerr:
            return False
        raise
    return True

@_lru_cache(maxsize=1)
def get_max_supported_compiler_dialect(clangpp):
    supported_dialects = []
    for dialect in range(30):
        if _check_dialect(dialect, clangpp):
            supported_dialects.append(dialect)

    if supported_dialects:
        return get_dialect_string(max(supported_dialects))
    else:
        return get_dialect_string(98)



'''
Target describes a single build or dependency target with all needed paths and
a list of buildables that comprise it's compile and link steps.
'''

import os as _os
from pathlib import Path as _Path
import subprocess as _subprocess
from multiprocessing import freeze_support as _freeze_support
import logging as _logging

from . import platform as _platform
from .dialect_check import get_dialect_string as _get_dialect_string
from .dialect_check import get_max_supported_compiler_dialect as _get_max_supported_compiler_dialect
from .build_type import BuildType as _BuildType
from .single_source import SingleSource as _SingleSource
from .progress_bar import get_build_progress_bar as _get_build_progress_bar

_LOGGER = _logging.getLogger('clang_build.clang_build')

class Target:
    DEFAULT_COMPILE_FLAGS          = ['-Wall', '-Werror']
    DEFAULT_RELEASE_COMPILE_FLAGS  = ['-O3', '-DNDEBUG']
    DEFAULT_DEBUG_COMPILE_FLAGS    = ['-O0', '-g3', '-DDEBUG']
    DEFAULT_COVERAGE_COMPILE_FLAGS = (
        DEFAULT_DEBUG_COMPILE_FLAGS +
        ['--coverage',
         '-fno-inline',
         '-fno-inline-small-functions',
         '-fno-default-inline'])


    def __init__(self,
            name,
            root_directory,
            build_directory,
            headers,
            include_directories,
            build_type,
            clangpp,
            options=None,
            dependencies=None):

        if options is None:
            options = {}
        if dependencies is None:
            dependencies = []

        self.dependency_targets = dependencies

        # Basics
        self.name           = name
        self.root_directory = _Path(root_directory)
        self.build_type      = build_type

        self.build_directory = build_directory

        self.headers = headers

        self.include_directories = []

        # Include directories
        if self.root_directory.joinpath('include').exists():
            self.include_directories.append(self.root_directory.joinpath('include'))
        self.include_directories += include_directories

        if 'properties' in options and 'cpp_version' in options['properties']:
            self.dialect = _get_dialect_string(options['properties']['cpp_version'])
        else:
            self.dialect = _get_max_supported_compiler_dialect(clangpp)

        # TODO: parse user-specified target version

        compile_flags        = []
        compile_flags_debug   = Target.DEFAULT_DEBUG_COMPILE_FLAGS
        compile_flags_release = Target.DEFAULT_RELEASE_COMPILE_FLAGS
        self.linkFlags = []

        if 'flags' in options:
            compile_flags += options['flags'].get('compile', [])
            compile_flags_release += options['flags'].get('compileRelease', [])
            compile_flags_debug += options['flags'].get('compileDebug', [])
            self.linkFlags += options['flags'].get('link', [])

        self.compile_flags = compile_flags
        if self.build_type == _BuildType.Release:
            self.compile_flags += compile_flags_release
        if self.build_type == _BuildType.Debug:
            self.compile_flags += compile_flags_debug

        for target in self.dependency_targets:
            self.compile_flags += target.compile_flags
            self.include_directories += target.include_directories
            self.headers += target.headers

        self.compile_flags = list(set(self.compile_flags))
        self.include_directories = list(set([dir.resolve() for dir in self.include_directories]))
        self.headers = list(set(self.headers))

        self.unsuccessful_builds = []

    def get_include_directory_command(self):
        ret = []
        for dir in self.include_directories:
            ret += ['-I',  str(dir.resolve())]
        return ret

    def link(self):
        # Subclasses must implement
        raise NotImplementedError()

    def compile(self, process_pool, progress_disabled):
        # Subclasses must implement
        raise NotImplementedError()

class HeaderOnly(Target):
    def link(self):
        _LOGGER.info(f'Header-only target [{self.name}] does not require linking.')

    def compile(self, process_pool, progress_disabled):
        _LOGGER.info(f'Header-only target [{self.name}] does not require compiling.')

def generate_depfile_single_source(buildable):
    buildable.generate_depfile()
    return buildable

def compile_single_source(buildable):
    buildable.compile()
    return buildable

class Compilable(Target):

    def __init__(self,
            name,
            root_directory,
            build_directory,
            headers,
            include_directories,
            source_files,
            build_type,
            clangpp,
            link_command,
            output_folder,
            platform_flags,
            prefix,
            suffix,
            options=None,
            dependencies=None):

        super().__init__(
            name=name,
            root_directory=root_directory,
            build_directory=build_directory,
            headers=headers,
            include_directories=include_directories,
            build_type=build_type,
            clangpp=clangpp,
            options=options,
            dependencies=dependencies)

        if not source_files:
            error_message = f'ERROR: Targt [{name}] was defined as a {self.__class__} but no source files were found'
            _LOGGER.error(error_message)
            raise RuntimeError(error_message)

        if options is None:
            options = {}
        if dependencies is None:
            dependencies = []

        self.object_directory  = self.build_directory.joinpath('obj').resolve()
        self.depfile_directory = self.build_directory.joinpath('dep').resolve()
        self.output_folder     = self.build_directory.joinpath(output_folder).resolve()

        self.object_directory.mkdir(parents=True, exist_ok=True)
        self.depfile_directory.mkdir(parents=True, exist_ok=True)
        self.output_folder.mkdir(parents=True, exist_ok=True)

        if 'output_name' in options:
            self.outname = options['output_name']
        else:
            self.outname = self.name

        self.outfile = _Path(self.output_folder, prefix + self.outname + suffix).resolve()


        # Clang
        self.clangpp   = clangpp

        # Sources
        self.source_files        = source_files

        # Buildables which this Target contains
        self.buildables = [_SingleSource(
            source_file=source_file,
            platform_flags=platform_flags,
            current_target_root_path=self.root_directory,
            depfile_directory=self.depfile_directory,
            object_directory=self.object_directory,
            include_strings=self.get_include_directory_command(),
            compile_flags=[self.dialect] + Target.DEFAULT_COMPILE_FLAGS + self.compile_flags,
            clangpp=self.clangpp) for source_file in self.source_files]

        # If compilation of buildables fail, they will be stored here later
        self.unsuccessful_builds = []

        # Linking setup
        self.link_command = link_command + [str(self.outfile)]

        self.link_command += self.linkFlags


        ## Additional scripts
        self.before_compile_script = ""
        self.before_link_script    = ""
        self.after_build_script    = ""
        if 'scripts' in options: ### TODO: maybe the scripts should be named differently
            self.before_compile_script = options['scripts'].get('before_compile', "")
            self.before_link_script    = options['scripts'].get('before_link',    "")
            self.after_build_script    = options['scripts'].get('after_build',    "")


    # From the list of source files, compile those which changed or whose dependencies (included headers, ...) changed
    def compile(self, process_pool, progress_disabled):

        # Object file only needs to be (re-)compiled if the source file or headers it depends on changed
        self.needed_buildables = [buildable for buildable in self.buildables if buildable.needs_rebuild]

        # If the target was not modified, it may not need to compile
        if not self.needed_buildables:
            _LOGGER.info(f'Target [{self.name}] is already compiled')
            return

        _LOGGER.info(f'Target [{self.name}] needs to build sources %s', [b.name for b in self.needed_buildables])

        # Before-compile step
        if self.before_compile_script:
            script_file = self.root_directory.joinpath(self.before_compile_script)
            _LOGGER.info(f'Pre-compile step of target [{self.name}]: {script_file}')
            original_directory = _os.getcwd()
            _os.chdir(self.root_directory)
            with open(script_file) as f:
                code = compile(f.read(), script_file, 'exec')
                exec(code, globals(), locals())
            _os.chdir(original_directory)
            _LOGGER.info(f'Finished pre-compile step of target [{self.name}]')

        # Execute depfile generation command
        _LOGGER.info(f'Scan dependencies of target [{self.outname}]')
        for b in self.needed_buildables:
            _LOGGER.debug(' '.join(b.dependency_command))
        self.needed_buildables = list(_get_build_progress_bar(
                process_pool.imap(
                    generate_depfile_single_source,
                    self.needed_buildables),
                progress_disabled,
                total=len(self.needed_buildables),
                name=self.name))

        # Execute compile command
        _LOGGER.info(f'Compile target [{self.outname}]')
        for b in self.needed_buildables:
            _LOGGER.debug(' '.join(b.compile_command))
        self.needed_buildables = list(_get_build_progress_bar(
                process_pool.imap(
                    compile_single_source,
                    self.needed_buildables),
                progress_disabled,
                total=len(self.needed_buildables),
                name=self.name))

        self.unsuccessful_builds = [buildable for buildable in self.needed_buildables if (buildable.compilation_failed or buildable.depfile_failed)]


    def link(self):
        # Before-link step
        if self.before_link_script:
            _LOGGER.info(f'Pre-link step of target [{self.name}]')
            original_directory = _os.getcwd()
            _os.chdir(self.root_directory)
            script_file = self.root_directory.joinpath(self.before_link_script)
            with open(script_file) as f:
                code = compile(f.read(), script_file, 'exec')
                exec(code, globals(), locals())
            _os.chdir(original_directory)
            _LOGGER.info(f'Finished pre-link step of target [{self.name}]')

        _LOGGER.info(f'Link target [{self.name}]')
        _LOGGER.debug('    ' + ' '.join(self.link_command))

        # Execute link command
        try:
            self.link_report = _subprocess.check_output(self.link_command, stderr=_subprocess.STDOUT).decode('utf-8').strip()
            self.unsuccessful_link = False
        except _subprocess.CalledProcessError as error:
            self.unsuccessful_link = True
            self.link_report = error.output.decode('utf-8').strip()

        ## After-build step
        if self.after_build_script:
            _LOGGER.info(f'After-build step of target [{self.name}]')
            original_directory = _os.getcwd()
            _os.chdir(self.root_directory)
            script_file = self.root_directory.joinpath(self.after_build_script)
            with open(script_file) as f:
                code = compile(f.read(), script_file, 'exec')
                exec(code, globals(), locals())
            _os.chdir(original_directory)
            _LOGGER.info(f'Finished after-build step of target [{self.name}]')


class Executable(Compilable):
    def __init__(self,
            name,
            root_directory,
            build_directory,
            headers,
            include_directories,
            source_files,
            build_type,
            clangpp,
            options=None,
            dependencies=None):

        super().__init__(
            name=name,
            root_directory=root_directory,
            build_directory=build_directory,
            headers=headers,
            include_directories=include_directories,
            source_files=source_files,
            build_type=build_type,
            clangpp=clangpp,
            link_command=[clangpp, '-o'],
            output_folder = _platform.EXECUTABLE_OUTPUT,
            platform_flags=_platform.PLATFORM_EXTRA_FLAGS_EXECUTABLE,
            prefix=_platform.EXECUTABLE_PREFIX,
            suffix=_platform.EXECUTABLE_SUFFIX,
            options=options,
            dependencies=dependencies)

        ### Library dependency search paths
        for target in self.dependency_targets:
            if not target.__class__ is HeaderOnly:
                self.link_command += ['-L', str(target.output_folder.resolve())]

        ### Link self
        self.link_command += [str(buildable.object_file) for buildable in self.buildables]

        ### Link dependencies
        for target in self.dependency_targets:
            if not target.__class__ is HeaderOnly:
                self.link_command += ['-l'+target.outname]


class SharedLibrary(Compilable):
    def __init__(self,
            name,
            root_directory,
            build_directory,
            headers,
            include_directories,
            source_files,
            build_type,
            clangpp,
            options=None,
            dependencies=None):

        super().__init__(
            name=name,
            root_directory=root_directory,
            build_directory=build_directory,
            headers=headers,
            include_directories=include_directories,
            source_files=source_files,
            build_type=build_type,
            clangpp=clangpp,
            link_command=[clangpp, '-shared', '-o'],
            output_folder = _platform.SHARED_LIBRARY_OUTPUT,
            platform_flags=_platform.PLATFORM_EXTRA_FLAGS_SHARED,
            prefix=_platform.SHARED_LIBRARY_PREFIX,
            suffix=_platform.SHARED_LIBRARY_SUFFIX,
            options=options,
            dependencies=dependencies)

        ### Library dependency search paths
        for target in self.dependency_targets:
            if not target.__class__ is HeaderOnly:
                self.link_command += ['-L', str(target.output_folder.resolve())]

        ### Link self
        self.link_command += [str(buildable.object_file) for buildable in self.buildables]

        ### Link dependencies
        for target in self.dependency_targets:
            if not target.__class__ is HeaderOnly:
                self.link_command += ['-l'+target.outname]


class StaticLibrary(Compilable):
    def __init__(self,
            name,
            root_directory,
            build_directory,
            headers,
            include_directories,
            source_files,
            build_type,
            clangpp,
            clang_ar,
            options=None,
            dependencies=None):

        super().__init__(
            name=name,
            root_directory=root_directory,
            build_directory=build_directory,
            headers=headers,
            include_directories=include_directories,
            source_files=source_files,
            build_type=build_type,
            clangpp=clangpp,
            link_command=[clang_ar, 'rc'],
            output_folder = _platform.STATIC_LIBRARY_OUTPUT,
            platform_flags=_platform.PLATFORM_EXTRA_FLAGS_STATIC,
            prefix=_platform.STATIC_LIBRARY_PREFIX,
            suffix=_platform.STATIC_LIBRARY_SUFFIX,
            options=options,
            dependencies=dependencies)

        # ### Include directories
        # self.link_command += self.get_include_directory_command()

        ### Link self
        self.link_command += [str(buildable.object_file) for buildable in self.buildables]

        ### Link dependencies
        for target in self.dependency_targets:
            if not target.__class__ is HeaderOnly:
                self.link_command += [str(buildable.object_file) for buildable in target.buildables]

if __name__ == '__main__':
    _freeze_support()



'''
Module containing custom errors that are
raised by clang-build if something goes wrong.
'''

class CompileError(RuntimeError):
    '''
    Error that is raised if compilation was
    not successful.
    '''
    def __init__(self, message, error_dict=None):
        '''
        :param message: Message of the error
        :param error_dict: A dict containing all errors
                           that occurred during compilation
        '''
        super().__init__(message)
        self.error_dict = error_dict

class LinkError(RuntimeError):
    '''
    Error that is raised if linking was
    not successful.
    '''
    def __init__(self, message, error_dict=None):
        '''
        :param message: Message of the error
        :param error_dict: A dict containing all errors
                           that occurred during compilation
        '''
        super().__init__(message)
        self.error_dict = error_dict


from tqdm import tqdm as _tqdm
from colorama import Fore as _Fore

_MAX_DESCRIPTION_WIDTH = 16
_ELLIPSIS_WIDTH = 2
_BAR_FORMAT = "{desc: <%s}: {percentage:3.0f}%% |%s{bar}%s| {n_fmt: >5}/{total_fmt: >5} [{elapsed: >8}]" % (_MAX_DESCRIPTION_WIDTH, _Fore.BLUE, _Fore.RESET)


def _format_lenghty_string(text):
    if len(text) > _MAX_DESCRIPTION_WIDTH:
        return ((text[:-(_MAX_DESCRIPTION_WIDTH-_ELLIPSIS_WIDTH)] and '.' * _ELLIPSIS_WIDTH)
                + text[-(_MAX_DESCRIPTION_WIDTH-_ELLIPSIS_WIDTH):])
    else:
        return text


def _get_clang_build_progress_bar(iterable, disable, total):
    return _tqdm(iterable, bar_format=_BAR_FORMAT, leave=False, disable=disable, total=total)


class CategoryProgress:
    def __init__(self, categories, disable, total=None):
        self.pbar = _get_clang_build_progress_bar(categories, disable, total)
        self.current_catgeory = 0
        self.pbar.set_description_str(categories[self.current_catgeory])
        self.categories = categories
        self.index_limit = len(self.categories)

    def __enter__(self):
        self.entered_pbar = self.pbar.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        self.entered_pbar.__exit__(type, value, traceback)

    def update(self):
        self.entered_pbar.update()
        self.current_catgeory += 1
        if self.current_catgeory >= self.index_limit:
            self.entered_pbar.set_description_str('Finished')
        else:
            self.entered_pbar.set_description_str(_format_lenghty_string(self.categories[self.current_catgeory]))

class IteratorProgress:
    def __init__(self, iterable, disable, total, access=lambda x: x):
        self.pbar = _get_clang_build_progress_bar(iterable, disable, total)
        self.access = access

    def __iter__(self):
        self.pbar_iter = self.pbar.__iter__()
        return self

    def __next__(self):
        val = self.pbar_iter.__next__()
        self.pbar.set_description_str(_format_lenghty_string(self.access(val)))
        return val

def get_build_progress_bar(iterable, disable, total, name):
    pbar = _get_clang_build_progress_bar(iterable, disable, total)
    pbar.set_description_str(_format_lenghty_string(name))
    return pbar



import os as _os
import re as _re
from pathlib import Path as _Path
import subprocess as _subprocess
from multiprocessing import freeze_support as _freeze_support
import logging as _logging

_LOGGER = _logging.getLogger('clang_build.clang_build')

# Find and parse the dependency file, return list of headers this file depends on
# See e.g. https://gcc.gnu.org/onlinedocs/gcc-8.1.0/gcc/Preprocessor-Options.html#Preprocessor-Options for documentation
# TODO: Can this be simplified?
def _get_depfile_headers(depfile):
    depfileHeaders = []
    with open(depfile, 'r') as the_file:
        depStr = the_file.read()
        # Find the first colon, which will be right after the object file name
        colonPos = depStr.find(':')
        # Separate the remainder into lines
        for line in depStr[colonPos + 1:].splitlines():
            # Remove the newline character ('\'-newline)
            if line.endswith('\\'):
                line = line[:-1]
            # Add header (or source, actually)
            depfileHeaders.append(_Path(line.strip().replace('\\ ', ' ')).resolve())
    return depfileHeaders

def _needs_rebuild(object_file, source_file, depfile):
    if depfile.exists():
        if object_file.exists():
            # If object file is found, check if it is up to date
            if source_file.stat().st_mtime > object_file.stat().st_mtime:
                return True
            # If object file is up to date, we check the headers it depends on
            else:
                for depHeaderFile in _get_depfile_headers(depfile):
                    if depHeaderFile.stat().st_mtime > object_file.stat().st_mtime:
                        return True

                return False
        else:
            return True
    else:
        return True

class SingleSource:
    def __init__(
            self,
            source_file,
            platform_flags,
            current_target_root_path,
            depfile_directory,
            object_directory,
            include_strings,
            compile_flags,
            clangpp):

        # Get the relative file path
        self.name          = source_file.name
        self.source_file    = source_file

        # If the source file is in a directory called 'src', we do not create a
        # subdirectory called 'src' in the build folder structure
        relpath = _os.path.relpath(source_file.parents[0], current_target_root_path)
        if current_target_root_path.joinpath('src').exists():
            relpath = _os.path.relpath(relpath, 'src')

        # Set name, extension and potentially produced output files
        self.object_file = _Path(object_directory,  relpath, self.source_file.stem + '.o')
        self.depfile     = _Path(depfile_directory, relpath, self.source_file.stem + '.d')


        self.needs_rebuild = _needs_rebuild(self.object_file, self.source_file, self.depfile)

        flags = compile_flags + include_strings

        self.compilation_failed = False

        # prepare everything for dependency file generation
        self.depfile.parents[0].mkdir(parents=True, exist_ok=True)
        self.dependency_command = [clangpp, '-E', '-MMD', str(self.source_file), '-MF', str(self.depfile)] + flags

        # prepare everything for compilation
        self.object_file.parents[0].mkdir(parents=True, exist_ok=True)
        self.compile_command = [clangpp, '-c', str(self.source_file), '-o', str(self.object_file)] + flags + platform_flags


    def generate_depfile(self):
        # TODO: logging in multiprocess
        # _LOGGER.debug('    ' + ' '.join(dependency_command))
        try:
            self.depfile_report = _subprocess.check_output(self.dependency_command, stderr=_subprocess.STDOUT).decode('utf-8').strip()
            self.depfile_failed = False
        except _subprocess.CalledProcessError as error:
            self.depfile_failed = True
            self.depfile_report = error.output.decode('utf-8').strip()


    def compile(self):
        # TODO: logging in multiprocess
        # _LOGGER.debug('    ' + ' '.join(self.compile_command))
        try:
            self.compile_report = _subprocess.check_output(self.compile_command, stderr=_subprocess.STDOUT).decode('utf-8').strip()
            self.compilation_failed = False
        except _subprocess.CalledProcessError as error:
            self.compilation_failed = True
            self.compile_report = error.output.decode('utf-8').strip()


if __name__ == '__name__':
    _freeze_support()


from sys import platform as _platform

if _platform == 'linux' or _platform == 'linux2':
    # Linux
    EXECUTABLE_PREFIX = ''
    EXECUTABLE_SUFFIX = ''
    EXECUTABLE_OUTPUT = 'bin'
    SHARED_LIBRARY_PREFIX = 'lib'
    SHARED_LIBRARY_SUFFIX = '.so'
    SHARED_LIBRARY_OUTPUT = 'lib'
    STATIC_LIBRARY_PREFIX = 'lib'
    STATIC_LIBRARY_SUFFIX = '.a'
    STATIC_LIBRARY_OUTPUT = 'lib'
    PLATFORM_EXTRA_FLAGS_EXECUTABLE = ['']
    PLATFORM_EXTRA_FLAGS_SHARED     = ['-fpic']
    PLATFORM_EXTRA_FLAGS_STATIC     = ['']
elif _platform == 'darwin':
    # OS X
    EXECUTABLE_PREFIX = ''
    EXECUTABLE_SUFFIX = ''
    EXECUTABLE_OUTPUT = 'bin'
    SHARED_LIBRARY_PREFIX = 'lib'
    SHARED_LIBRARY_SUFFIX = '.dylib'
    SHARED_LIBRARY_OUTPUT = 'lib'
    STATIC_LIBRARY_PREFIX = 'lib'
    STATIC_LIBRARY_SUFFIX = '.a'
    STATIC_LIBRARY_OUTPUT = 'lib'
    PLATFORM_EXTRA_FLAGS_EXECUTABLE = ['']
    PLATFORM_EXTRA_FLAGS_SHARED     = ['']
    PLATFORM_EXTRA_FLAGS_STATIC     = ['']
elif _platform == 'win32':
    # Windows
    EXECUTABLE_PREFIX = ''
    EXECUTABLE_SUFFIX = '.exe'
    EXECUTABLE_OUTPUT = 'bin'
    SHARED_LIBRARY_PREFIX = ''
    SHARED_LIBRARY_SUFFIX = '.dll'
    SHARED_LIBRARY_OUTPUT = 'bin'
    STATIC_LIBRARY_PREFIX = ''
    STATIC_LIBRARY_SUFFIX = '.lib'
    STATIC_LIBRARY_OUTPUT = 'lib'
    PLATFORM_EXTRA_FLAGS_EXECUTABLE = ['-Xclang', '-flto-visibility-public-std']
    PLATFORM_EXTRA_FLAGS_SHARED     = ['-Xclang', '-flto-visibility-public-std']
    PLATFORM_EXTRA_FLAGS_STATIC     = ['-Xclang', '-flto-visibility-public-std']
else:
    raise RuntimeError('Platform ' + _platform + 'is currently not supported.')


from glob import iglob as _iglob
from pathlib import Path as _Path

def _get_header_files(folder, recursive=True):
    headers = []
    for ext in ('*.hpp', '*.hxx', '*.h'):
        headers += [_Path(f) for f in _iglob(str(folder) + '/**/'+ext, recursive=recursive)]

    return headers

def _get_source_files(folder, recursive=True):
    sources = []
    if recursive:
        for ext in ('*.cpp', '*.cxx', '*.c'):
            sources += [_Path(f) for f in _iglob(str(folder) + '/**/'+ext, recursive=recursive)]
    else:
        for ext in ('*.cpp', '*.cxx', '*.c'):
            sources += [_Path(f) for f in _iglob(str(folder) + '/*'+ext, recursive=recursive)]

    return sources

def get_sources_and_headers(target_options, target_root_directory, target_build_directory):
    output = {'headers': [], 'include_directories': [], 'sourcefiles': []}
    relative_includes = []
    relative_source_directories = []

    # TODO: maybe the output should also include the root dir, build dir and potentially download dir?
    # TODO: should warn when a specified directory does not exist!

    # Find source files
    headers_specified = False
    if 'sources' in target_options:
        sourcenode = target_options['sources']
        if 'include_directories' in sourcenode:
            headers_specified = True
            output['include_directories'] += list(set(target_root_directory.joinpath(path) for path in sourcenode['include_directories']))
            for directory in output['include_directories']:
                output['headers'] += _get_header_files(directory, recursive=True)
    if not headers_specified:
        output['include_directories'] += [target_root_directory.joinpath(''), target_root_directory.joinpath('include'), target_root_directory.joinpath('thirdparty')]
        for directory in output['include_directories']:
            output['headers'] += _get_header_files(target_root_directory.joinpath(directory), recursive=False)



    sources_specified = False
    if 'sources' in target_options:
        sourcenode = target_options['sources']
        if 'source_directories' in sourcenode:
            sources_specified = True
            for directory in list(set(target_root_directory.joinpath(path) for path in sourcenode['source_directories'])):
                output['sourcefiles'] += _get_source_files(directory, recursive=True)
    if not sources_specified:
        output['sourcefiles'] += _get_source_files(target_root_directory.joinpath('src'), recursive=True)
    if not output['sourcefiles']:
        output['sourcefiles'] += _get_source_files(target_root_directory, recursive=False)


    output['include_directories'] = list(set(output['include_directories']))
    output['headers']             = list(set(output['headers']))
    output['sourcefiles']         = list(set(output['sourcefiles']))

    return output